/*****************************************************************************\
 *  builtin.c - Simple builtin (FIFO) scheduler plugin.
 *		Periodically when pending jobs can start.
 *		This is a minimal implementation of the logic found in
 *		src/plugins/sched/backfill/backfill.c and disregards
 *		how jobs are scheduled sequencially.
 *****************************************************************************
 *  Copyright (C) 2003-2007 The Regents of the University of California.
 *  Copyright (C) 2008-2010 Lawrence Livermore National Security.
 *  Produced at Lawrence Livermore National Laboratory (cf, DISCLAIMER).
 *  Written by Morris Jette <jette1@llnl.gov>
 *  CODE-OCEC-09-009. All rights reserved.
 *
 *  This file is part of Slurm, a resource management program.
 *  For details, see <https://slurm.schedmd.com/>.
 *  Please also read the included file: DISCLAIMER.
 *
 *  Slurm is free software; you can redistribute it and/or modify it under
 *  the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  In addition, as a special exception, the copyright holders give permission
 *  to link the code of portions of this program with the OpenSSL library under
 *  certain conditions as described in each individual source file, and
 *  distribute linked combinations including the two. You must obey the GNU
 *  General Public License in all respects for all of the code used other than
 *  OpenSSL. If you modify file(s) with this exception, you may extend this
 *  exception to your version of the file(s), but you are not obligated to do
 *  so. If you do not wish to do so, delete this exception statement from your
 *  version.  If you delete this exception statement from all source files in
 *  the program, then also delete it here.
 *
 *  Slurm is distributed in the hope that it will be useful, but WITHOUT ANY
 *  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 *  details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with Slurm; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA.
\*****************************************************************************/

#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "slurm/slurm.h"
#include "slurm/slurm_errno.h"

#include "src/common/list.h"
#include "src/common/macros.h"
#include "src/common/node_select.h"
#include "src/common/parse_time.h"
#include "src/common/slurm_protocol_api.h"
#include "src/common/xmalloc.h"
#include "src/common/xstring.h"

#include "src/slurmctld/burst_buffer.h"
#include "src/slurmctld/locks.h"
#include "src/slurmctld/preempt.h"
#include "src/slurmctld/reservation.h"
#include "src/slurmctld/slurmctld.h"
#include "src/plugins/sched/builtin/builtin.h"

// copy pasta of the backfill scheduler
#ifdef SLURM_SIMULATOR
#include <fcntl.h>           /* For O_* constants */
#include <sys/stat.h>        /* For mode constants */
#include <semaphore.h>
#include <pthread.h>
#include "src/common/slurm_sim.h"
#endif

#ifndef BACKFILL_INTERVAL
#  define BACKFILL_INTERVAL	30
#endif


/*********************** local variables *********************/
static bool stop_builtin = false;
static pthread_mutex_t term_lock = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t  term_cond = PTHREAD_COND_INITIALIZER;
static pthread_mutex_t config_lock = PTHREAD_MUTEX_INITIALIZER;
static bool config_flag = false;
static int builtin_interval = BACKFILL_INTERVAL;
static int max_sched_job_cnt = 50;
static int sched_timeout = 0;
// copy pasta of the backfill algorithm
#ifndef SLURM_SIMULATOR
static int backfill_interval = BACKFILL_INTERVAL; // if it is simulator backfill interval will be a global variable
char SEM_NAME[]         = "serversem";
sem_t* mutexserver      = SEM_FAILED;
#endif

/*********************** local functions *********************/
static void _compute_start_times(void);
static void _load_config(void);
#ifndef SLURM_SIMULATOR
static void _my_sleep(int secs);
#else
static uint32_t _my_sleep(int64_t usec);

#endif

/* Terminate builtin_agent */
extern void stop_builtin_agent(void)
{
	slurm_mutex_lock(&term_lock);
	stop_builtin = true;
	slurm_cond_signal(&term_cond);
	slurm_mutex_unlock(&term_lock);
}


#ifndef SLURM_SIMULATOR
static void _my_sleep(int secs)
{
	struct timespec ts = {0, 0};
	struct timeval now;

	gettimeofday(&now, NULL);
	ts.tv_sec = now.tv_sec + secs;
	ts.tv_nsec = now.tv_usec * 1000;
	slurm_mutex_lock(&term_lock);
	if (!stop_builtin)
		slurm_cond_timedwait(&term_cond, &term_lock, &ts);
	slurm_mutex_unlock(&term_lock);
}
#else
// replaced the _my_sleep function with backfill's one
/* Sleep for at least specified time, returns actual sleep time in usec */
static uint32_t _my_sleep(int64_t usec)
{
        int64_t nsec;
        uint32_t sleep_time = 0;
        struct timespec ts = {0, 0};
        struct timeval  tv1 = {0, 0}, tv2 = {0, 0};

        if (gettimeofday(&tv1, NULL)) {         /* Some error */
                sleep(1);
                return 1000000;
        }

        nsec  = tv1.tv_usec + usec;
        nsec *= 1000;
        ts.tv_sec  = tv1.tv_sec + (nsec / 1000000000);
        ts.tv_nsec = nsec % 1000000000;
        slurm_mutex_lock(&term_lock);
        if (!stop_builtin)
                slurm_cond_timedwait(&term_cond, &term_lock, &ts);
        slurm_mutex_unlock(&term_lock);
        if (gettimeofday(&tv2, NULL))
                return usec;
        sleep_time = (tv2.tv_sec - tv1.tv_sec) * 1000000;
        sleep_time += tv2.tv_usec;
        sleep_time -= tv1.tv_usec;
        return sleep_time;
}
#endif

// super copy pasta of the backfill algorithm
/*
 * _many_pending_rpcs - Determine if slurmctld is busy with many active RPCs
 * RET - True if slurmctld currently has more than SLURMCTLD_THREAD_LIMIT
 *       active RPCs
 */
static bool _many_pending_rpcs(void)
{
#ifndef SLURM_SIMULATOR
        bool many_pending_rpcs = false;

        slurm_mutex_lock(&slurmctld_config.thread_count_lock);
        //info("thread_count = %u", slurmctld_config.server_thread_count);
        if ((max_rpc_cnt > 0) &&
            (slurmctld_config.server_thread_count >= max_rpc_cnt))
                many_pending_rpcs = true;
        slurm_mutex_unlock(&slurmctld_config.thread_count_lock);

        return many_pending_rpcs;
#else
        /* Marco: Simulator: time should be not passing here so no rpc should arrive */
        return false;
#endif
}

static void _load_config(void)
{
	char *sched_params = slurm_get_sched_params();
	char *tmp_ptr;

	sched_timeout = slurm_get_msg_timeout() / 2;
	sched_timeout = MAX(sched_timeout, 1);
	sched_timeout = MIN(sched_timeout, 10);

	if ((tmp_ptr = xstrcasestr(sched_params, "interval=")))
		builtin_interval = atoi(tmp_ptr + 9);
	if (builtin_interval < 1) {
		error("Invalid SchedulerParameters interval: %d",
		      builtin_interval);
		builtin_interval = BACKFILL_INTERVAL;
	}

	if ((tmp_ptr = xstrcasestr(sched_params, "max_job_bf=")))
		max_sched_job_cnt = atoi(tmp_ptr + 11);
	if ((tmp_ptr = xstrcasestr(sched_params, "bf_max_job_test=")))
		max_sched_job_cnt = atoi(tmp_ptr + 16);
	if (max_sched_job_cnt < 1) {
		error("Invalid SchedulerParameters bf_max_job_test: %d",
		      max_sched_job_cnt);
		max_sched_job_cnt = 50;
	}
	xfree(sched_params);
}

static void _compute_start_times(void)
{
	int j, rc = SLURM_SUCCESS, job_cnt = 0;
	List job_queue;
	job_queue_rec_t *job_queue_rec;
	List preemptee_candidates = NULL;
	struct job_record *job_ptr;
	struct part_record *part_ptr;
	bitstr_t *alloc_bitmap = NULL, *avail_bitmap = NULL;
	bitstr_t *exc_core_bitmap = NULL;
	uint32_t max_nodes, min_nodes, req_nodes, time_limit;
	time_t now = time(NULL), sched_start, last_job_alloc;
	bool resv_overlap = false;

	sched_start = now;
	last_job_alloc = now - 1;
	alloc_bitmap = bit_alloc(node_record_count);
	job_queue = build_job_queue(true, false);
	sort_job_queue(job_queue);
	while ((job_queue_rec = (job_queue_rec_t *) list_pop(job_queue))) {
		job_ptr  = job_queue_rec->job_ptr;
		part_ptr = job_queue_rec->part_ptr;
		xfree(job_queue_rec);
		if (part_ptr != job_ptr->part_ptr)
			continue;	/* Only test one partition */

		if (job_cnt++ > max_sched_job_cnt) {
			debug2("scheduling loop exiting after %d jobs",
			       max_sched_job_cnt);
			break;
		}

		/* Determine minimum and maximum node counts */
		/* On BlueGene systems don't adjust the min/max node limits
		   here.  We are working on midplane values. */
		min_nodes = MAX(job_ptr->details->min_nodes,
				part_ptr->min_nodes);

		if (job_ptr->details->max_nodes == 0)
			max_nodes = part_ptr->max_nodes;
		else
			max_nodes = MIN(job_ptr->details->max_nodes,
					part_ptr->max_nodes);

		max_nodes = MIN(max_nodes, 500000);     /* prevent overflows */

		if (job_ptr->details->max_nodes)
			req_nodes = max_nodes;
		else
			req_nodes = min_nodes;

		if (min_nodes > max_nodes) {
			/* job's min_nodes exceeds partition's max_nodes */
			continue;
		}

		j = job_test_resv(job_ptr, &now, true, &avail_bitmap,
				  &exc_core_bitmap, &resv_overlap, false);
		if (j != SLURM_SUCCESS) {
			FREE_NULL_BITMAP(avail_bitmap);
			FREE_NULL_BITMAP(exc_core_bitmap);
			continue;
		}

		rc = select_g_job_test(job_ptr, avail_bitmap,
				       min_nodes, max_nodes, req_nodes,
				       SELECT_MODE_WILL_RUN,
				       preemptee_candidates, NULL,
				       exc_core_bitmap);
		if (rc == SLURM_SUCCESS) {
			last_job_update = now;
			if (job_ptr->time_limit == INFINITE)
				time_limit = 365 * 24 * 60 * 60;
			else if (job_ptr->time_limit != NO_VAL)
				time_limit = job_ptr->time_limit * 60;
			else if (job_ptr->part_ptr &&
				 (job_ptr->part_ptr->max_time != INFINITE))
				time_limit = job_ptr->part_ptr->max_time * 60;
			else
				time_limit = 365 * 24 * 60 * 60;
			if (bit_overlap(alloc_bitmap, avail_bitmap) &&
			    (job_ptr->start_time <= last_job_alloc)) {
				job_ptr->start_time = last_job_alloc;
			}
			bit_or(alloc_bitmap, avail_bitmap);
			last_job_alloc = job_ptr->start_time + time_limit;
		}
		FREE_NULL_BITMAP(avail_bitmap);
		FREE_NULL_BITMAP(exc_core_bitmap);

		if ((time(NULL) - sched_start) >= sched_timeout) {
			debug2("scheduling loop exiting after %d jobs",
			       max_sched_job_cnt);
			break;
		}
	}
	FREE_NULL_LIST(job_queue);
	FREE_NULL_BITMAP(alloc_bitmap);
}

/* Note that slurm.conf has changed */
extern void builtin_reconfig(void)
{
	slurm_mutex_lock(&config_lock);
	config_flag = true;
	slurm_mutex_unlock(&config_lock);
}

#ifdef SLURM_SIMULATOR

char BI_SEM_NAME[] = "bi_sem";
char BI_DONE_SEM_NAME[] = "bi_done_sem";
sem_t* mutex_bi_pg=NULL;
sem_t* mutex_bi_done_pg=NULL;

int open_BI_sync_semaphore_pg() {
        mutex_bi_pg = sem_open(BI_SEM_NAME, O_CREAT, 0644, 0);
        if(mutex_bi_pg == SEM_FAILED) {
                error("unable to create builtin semaphore");
                sem_unlink(BI_SEM_NAME);
                return -1;
        }

        mutex_bi_done_pg = sem_open(BI_DONE_SEM_NAME, O_CREAT, 0644, 0);
        if(mutex_bi_done_pg == SEM_FAILED) {
                error("unable to create backfill done semaphore");
                sem_unlink(BI_DONE_SEM_NAME);
                return -1;
        }

        return 0;
}

void close_BI_sync_semaphore() {
        if(mutex_bi_pg != SEM_FAILED) sem_close(mutex_bi_pg);
        if(mutex_bi_done_pg != SEM_FAILED) sem_close(mutex_bi_done_pg);
}

#endif

/* builtin_agent - detached thread periodically when pending jobs can start */
extern void *builtin_agent(void *args)
{
	time_t now;
	double wait_time;
	static time_t last_sched_time = 0;

	/* Read config, nodes and partitions; Write jobs */
	slurmctld_lock_t all_locks = {
		READ_LOCK, WRITE_LOCK, READ_LOCK, READ_LOCK, READ_LOCK };

	// this seems to be some debugging flags... 
	bool load_config;
        bool short_sleep = false;
	_load_config();
        open_BI_sync_semaphore_pg();

	while (!stop_builtin) {
                sem_wait(mutex_bi_pg);
		_my_sleep((int64_t) builtin_interval * 1000000);
		if (stop_builtin){
			sem_wait(mutex_bi_done_pg);
			break;
		}
                slurm_mutex_lock(&config_lock);
                if (config_flag) {
                        config_flag = false;
                        load_config = true;
                } else {
                        load_config = false;
                }
                slurm_mutex_unlock(&config_lock);

		if (load_config){
                        _load_config();
                }

		now = time(NULL);
		wait_time = difftime(now, last_sched_time);

                if ((wait_time < builtin_interval) ||
                    job_is_completing(NULL) || _many_pending_rpcs() ||
                    !avail_front_end(NULL) || !_more_work(last_sched_time)) {
                        short_sleep = true;
                        continue;
                }

		slurm_mutex_lock(&check_bf_running_lock);
		slurmctld_diag_stats.bf_active = 1;
                slurm_mutex_unlock(&check_bf_running_lock);


                debug("builtin: now %ld, last_builtin_time %ld, wait_time %lf, ", now, last_sched_time, wait_time);
                if (!(job_is_completing(NULL) || _many_pending_rpcs() ||
                         !avail_front_end(NULL) || !_more_work(last_sched_time))) {
			lock_slurmctld(all_locks);
			_compute_start_times();
			last_sched_time = time(NULL);
			(void) bb_g_job_try_stage_in();
			unlock_slurmctld(all_locks);		
		}

		slurm_mutex_lock(&check_bf_running_lock);
		slurmctld_diag_stats.bf_active = 1;
                slurm_mutex_unlock(&check_bf_running_lock);

		short_sleep = false;
		sem_post(mutex_bi_done_pg);
	}
	close_BI_sync_semaphore();
	return NULL;
}


/*
 * Report summary of job's feature specification
 * IN job_ptr - job to schedule
 * OUT has_xand - true if features are XANDed together
 * OUT has_xor - true if features are XORed together
 * RET Total count for ALL job features, even counts with XAND separator
 */
static int _num_feature_count(struct job_record *job_ptr, bool *has_xand,
                              bool *has_xor)
{
        struct job_details *detail_ptr = job_ptr->details;
        int rc = 0;
        ListIterator feat_iter;
        job_feature_t *feat_ptr;

        *has_xand = false;
        *has_xor = false;
        if (detail_ptr->feature_list == NULL)   /* no constraints */
                return rc;

        feat_iter = list_iterator_create(detail_ptr->feature_list);
        while ((feat_ptr = (job_feature_t *) list_next(feat_iter))) {
                if (feat_ptr->count)
                        rc++;
                if (feat_ptr->op_code == FEATURE_OP_XAND)
                        *has_xand = true;
                if (feat_ptr->op_code == FEATURE_OP_XOR)
                        *has_xor = true;
        }
        list_iterator_destroy(feat_iter);

        return rc;
}

